(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SubtermCriterionImpl imports SubtermCriterion DpTransImpl begin

datatype 'f spL = SimpProj "('f \<times> nat)list"

definition check_supt_proj :: "'f sp \<Rightarrow> ('f::shows,'v::shows)rule \<Rightarrow> shows monad"
where "check_supt_proj p lr \<equiv> checkB(checkSupterm (lift2term p (fst lr)) (lift2term p (snd lr)))
        (''the projected rhs is not a proper subterm of the projected lhs'' +#+ shows_nl)
      \<leftarrow> (\<lambda>s. ''Could not orient rule '' +#+ shows_rule lr +@+ shows_nl +@+
             ''i.e., '' +#+ shows(lift2term p (fst lr)) +@+ '' |> '' +#+ shows(lift2term p (snd lr)) +@+ shows_nl +@+ s)
"

definition check_supteq_proj ::
  "'f sp \<Rightarrow> ('f::shows,'v::shows)rule \<Rightarrow> shows monad"
where "check_supteq_proj p lr \<equiv> checkB(lift2term p (fst lr) = lift2term p (snd lr)
       \<or> checkSupterm (lift2term p (fst lr)) (lift2term p (snd lr)))
         (''the projected rhs is not a subterm of the projected lhs'' +#+ shows_nl)
      \<leftarrow> (\<lambda>s. ''Could not orient rule '' +#+ shows lr +@+ shows_nl +@+
             ''i.e., '' +#+ shows(lift2term p (fst lr)) +@+ '' |>= '' +#+ shows(lift2term p (snd lr)) +@+ shows_nl +@+ s)
"

fun check_projection :: "('f \<Rightarrow> ('f,'v)trsL) \<Rightarrow> 'f sp \<Rightarrow> ('f::shows,'v::shows)rule \<Rightarrow> shows monad"
where "check_projection ds p (Fun f ss,Fun g ts) = (
  let ls = length ss; lt = length ts in (
  checkB(0 < ls) (''lhs is a constant'' +#+ shows_nl)
\<guillemotright> checkB(p f < ls) (shows(p f) +@+ '' is not an argument position of '' +#+ shows(Fun f ss) +@+ shows_nl)

\<guillemotright>  checkB(0 < lt) (''rhs is a constant'' +#+ shows_nl)
\<guillemotright> checkB(ds g = []) (shows g +@+ '' is defined in R'' +#+ shows_nl)
\<guillemotright> checkB(p g < lt) (shows(p g) +@+ '' is not an argument position of '' +#+ shows(Fun g ts) +@+ shows_nl)
))"
    | "check_projection ds p (s,t) = error(''cannot project a variable'' +#+ shows_nl)"

lemma check_supt_proj_sound[simp]:
  assumes "isOK(check_supt_proj p r)"
  shows "(lift2term p (fst r),lift2term p (snd r)) \<in> supt"
proof -
  from assms have "checkSupterm (lift2term p (fst r)) (lift2term p (snd r))"
    unfolding check_supt_proj_def catchSound3 checkBSound .
  thus ?thesis unfolding checkSuptermSound[to_set] by simp
qed

lemma check_supteq_proj_sound[simp]:
  assumes "isOK(check_supteq_proj p r)"
  shows "(lift2term p (fst r),lift2term p (snd r)) \<in> supteq" (is "(?s,?t) \<in> _")
proof -
  from assms have "?s = ?t \<or> checkSupterm ?s ?t"
    unfolding check_supteq_proj_def catchSound3 checkBSound .
  thus ?thesis unfolding checkSuptermSound[to_set] by (rule disjE) auto
qed

lemma check_projection_sound[simp]:
  assumes check: "isOK(check_projection d p rule)"
  and map: "computeRuleMap R = Some d"
  shows "(\<exists>f ss. fst rule = Fun f ss \<and> 0 < length ss \<and> p f < length ss)
       \<and> (\<exists>g ts. snd rule = Fun g ts \<and> 0 < length ts \<and> g \<notin> defs(set R) \<and> p g < length ts)"
proof -
  show ?thesis
  proof (cases rule)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var x) from assms show ?thesis unfolding Pair Var by simp
    next
      case (Fun f ss) 
      hence Fun1: "l = Fun f ss" .
      show ?thesis
      proof (cases r)
        case (Var y) from assms show ?thesis unfolding Pair Var by simp
      next
        case (Fun g ts)        
        with check computeRuleMapSound[OF map, of g] 
        show ?thesis unfolding Pair Fun1 Fun by  (auto simp: defs_def defined_def)
      qed
    qed
  qed
qed


(*
fun check_spsc_proc ::
  "'f sp \<Rightarrow> sstring list \<Rightarrow>
    ('f::sshow,'v::sshow)trsL \<Rightarrow>
    ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow> sstring list monad"
where "check_spsc_proc proj rep P rP R = (
  (let ds = computeDSymbs R in (
    checkAllM (check_projection ds proj) P
    \<leftarrow> (\<lambda>s. ''the supplied projection is malformed'' # s))
  )
\<guillemotright> (let (Sp::('f,'v)trsL,NSp) = List.partition (\<lambda>r. isOK(check_supt_proj proj r)) P in
  (
    checkAllM (check_supteq_proj proj) NSp
    \<leftarrow> (\<lambda>s. ''problem when orienting DPs'' # s)
  \<guillemotright> checkAllM check_for_var R
    \<leftarrow> (\<lambda>s. ''one of the lhss of R is a variable'' # s)
  \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda>s. case check_supt_proj proj s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'',''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
  ))
\<leftarrow> (\<lambda> s. (''could not apply the subterm criterion processor with the following simple projection''
  # (rep@([]#s))))
)"
*)

fun check_spsc_proc ::
  "'f sp \<Rightarrow> shows \<Rightarrow>
    ('f::shows,'v::shows)trsL \<Rightarrow>
    ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> shows monad"
where "check_spsc_proc proj rep P rP R (Some rm) = (
  let (Sp::('f,'v)trsL,NSp) = List.partition (\<lambda>r. isOK(check_supt_proj proj r)) P in (
    checkAllM (check_projection rm proj) P
    \<leftarrow> (\<lambda>s. ''the supplied projection is malformed'' +#+ shows_nl +@+ s)
  \<guillemotright> checkAllM (check_supteq_proj proj) NSp
    \<leftarrow> (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
  \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda>s. case check_supt_proj proj s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
)
\<leftarrow> (\<lambda> s. ''could not apply the subterm criterion processor with the following simple projection'' +#+ shows_nl +@+
        rep +@+ shows_nl +@+ s)
)"
| "check_spsc_proc proj rep P rP R None = OK"

lemma check_spsc_proc_sound[simp] :
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(check_spsc_proc proj rep P rP R rm))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof ((rule allI)+, intro impI)
  fix P R rP
  let ?\<pi> = "lift2term proj"
  assume check: "isOK(check_spsc_proc proj rep P rP R (computeRuleMap R))"
  show "\<forall>s t \<sigma>. min_ichain (set P,set R) s t \<sigma> \<longrightarrow> (\<exists>i. ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (cases "computeRuleMap R")
    case (Some rm)
    have "subterm_criterion_proc proj (set P) (set R) (set rP) (set R)"
    proof (cases "List.partition (\<lambda>dp. isOK(check_supt_proj proj dp)) P")
      case (Pair Sp NSp)
      have "spscCond proj (set P) (set rP) (set R)"
      proof -
        from Pair have P: "set P = set NSp \<union> set Sp" by auto
        from check Some Pair have subset: "set NSp \<subseteq> set rP" by auto
        with P have Sp: "set P - set rP \<subseteq> set Sp" by auto
        
        from check Some Pair have "\<forall>lr\<in>set NSp. isOK(check_supteq_proj proj lr)" by (auto simp: Let_def mem_iff)
        with check_supteq_proj_sound[where p = proj]
        have nonstrict: "\<forall>lr\<in>set NSp. (?\<pi>(fst lr),?\<pi>(snd lr)) \<in> supteq" by best
        
        from check Pair have strict: "\<forall>lr\<in>set Sp. isOK(check_supt_proj proj lr)" by (auto simp: Let_def mem_iff)
        
        with check_supt_proj_sound[where p = proj]
        have "\<forall>lr\<in>set Sp. (?\<pi>(fst lr),?\<pi>(snd lr)) \<in> supt" by best
        hence "\<forall>lr\<in>set Sp. (?\<pi>(fst lr),?\<pi>(snd lr)) \<in> supteq" unfolding supt_iff_supteq_neq[to_set] by auto
        with P nonstrict
        have supteq: "\<forall>lr\<in>set P. (?\<pi>(fst lr),?\<pi>(snd lr)) \<in> supteq"
          (is "\<forall>lr\<in>set P. ?Psupteq (fst lr) (snd lr)") by best
        
        from Sp strict have "\<forall>lr\<in>set P - set rP. isOK(check_supt_proj proj lr)" by auto
        with check_supt_proj_sound[where p = proj]
        have "\<forall>lr\<in>set P - set rP. (?\<pi>(fst lr),?\<pi>(snd lr)) \<in> supt" by best
        hence A: "\<forall>(l,r)\<in>set P - set rP. (?\<pi> l,?\<pi> r) \<in> supt" unfolding split_def by auto
        
        from check Some have "\<forall>lr\<in>set P. isOK(check_projection rm proj lr)" by auto
        hence "\<forall>lr\<in>set P. (\<exists>f ss. fst lr = Fun f ss \<and> 0 < length ss \<and> proj f < length ss)
          \<and>  (\<exists>g ts. snd lr = Fun g ts \<and> 0 < length ts \<and> g \<notin> defs(set R) \<and> proj g < length ts)"
          (is "\<forall>lr\<in>set P. ?P (fst lr) (snd lr)")
          using check_projection_sound[OF _ Some, where p = proj] by blast
        with supteq have "\<forall>lr\<in>set P. ?Psupteq (fst lr) (snd lr) \<and> ?P (fst lr) (snd lr)" by auto
        hence B: "\<forall>(l,r)\<in>set P. ?Psupteq l r \<and> ?P l r" unfolding split_def by auto
        from A and B show ?thesis unfolding spscCond_def by best
        
      qed
      moreover have "set R = set R" by simp
      ultimately show ?thesis unfolding subterm_criterion_proc_def by auto
    qed
    with ci_subterm_criterion_proc
    show ?thesis unfolding chain_identifying_proc_def by best
  next
    case None
    then obtain l r where rule: "(l,r) \<in> set R" and "root l = None" by (simp only: computeRuleMapSound2, auto)
    then obtain x where xr: "(Var x, r) \<in> set R" by (cases l, auto)
    have notSN: "\<And> t. \<not> SN_elt (rstep (set R)) t" using rule left_var_imp_not_SN[OF xr] by auto
    show ?thesis
    proof (intro allI impI)
      fix s t \<sigma>
      assume "min_ichain (set P, set R) s t \<sigma>"
      hence "SN_elt (rstep (set R)) (t 0 \<cdot> \<sigma> 0)" by (simp add: minimal_cond_def)
      with notSN have False ..
      thus "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" ..
    qed
  qed
qed

fun create_sp :: "'f spL \<Rightarrow> 'f sp"
where "create_sp (SimpProj sp) f = (case lookup f sp of None \<Rightarrow> 0 | Some n \<Rightarrow> n)"

fun create_sp_representation :: "('f::shows)spL \<Rightarrow> shows"
where "create_sp_representation(SimpProj sp) = (shows_map (\<lambda>(f,n).
  ''pi('' +#+ shows f +@+ '') = '' +#+ shows n +@+ shows_nl
) sp)"
  
fun create_spsc_proc :: "'f spL \<Rightarrow> ('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL) option \<Rightarrow> shows monad"
where "create_spsc_proc sp P rP R rm = (let \<pi> = create_sp sp in
  check_spsc_proc \<pi> (create_sp_representation sp) P rP R rm)"

lemma spsc_proc_ci:
  "chain_identifying_onlyDP_proc_impl(\<lambda>P R rm rrm rP. isOK(create_spsc_proc sp P rP R rm))"
proof -
  let ?sp = "create_sp sp"
  have simpme: "create_spsc_proc sp = check_spsc_proc ?sp (create_sp_representation sp)" by (rule ext, rule ext, rule ext, rule ext, simp add: Let_def)
  show ?thesis by (simp only: simpme, rule check_spsc_proc_sound)
qed

declare create_spsc_proc.simps[simp del]

end
